"""Module for model pipeline and pipenet.

Pipeline or Pipenet connects multiple models (transformers, detectors, and/or
aggregators) into a "super" model that may perform complex anomaly detection
process.

"""
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.collections import PatchCollection
from matplotlib.lines import Line2D
from matplotlib.patches import Circle
from tabulate import tabulate

from .._aggregator_base import _Aggregator
from .._base import _Model, _TrainableModel
from .._detector_base import (  # _NonTrainableMultivariateDetector,
    _NonTrainableUnivariateDetector,
    _TrainableMultivariateDetector,
    _TrainableUnivariateDetector,
)
from .._transformer_base import (
    _NonTrainableMultivariateTransformer,
    _NonTrainableUnivariateTransformer,
    _TrainableMultivariateTransformer,
    _TrainableUnivariateTransformer,
)
from ..metrics import f1_score, iou, precision, recall

_Detector = (
    _NonTrainableUnivariateDetector,
    # _NonTrainableMultivariateDetector,
    _TrainableUnivariateDetector,
    _TrainableMultivariateDetector,
)
_Transformer = (
    _NonTrainableUnivariateTransformer,
    _NonTrainableMultivariateTransformer,
    _TrainableUnivariateTransformer,
    _TrainableMultivariateTransformer,
)


class Pipeline:
    """A Pipeline object chains transformers and a detector sequentially.

    Parameters
    ----------
    steps: list of 2-tuples (str, object)
        Components of this pipeline. Each 2-tuple represents a step in the
        pipeline (step name, model object).

    Examples
    --------
    >>> steps = [('moving average', RollingAggregate(agg='mean', window=10)),
                 ('filter quantile 0.99', QuantileAD(high=0.99))]
    >>> myPipeline = Pipeline(steps)

    """

    def __init__(self, steps: List[Tuple[str, _Model]]) -> None:
        self.steps = steps
        self._pipenet = Pipenet()
        self._update_internal_pipenet()

    def _update_internal_pipenet(self) -> None:
        pipenet_steps = dict()
        last_name = "original"
        for pipeline_step in self.steps:
            pipenet_steps.update(
                {
                    pipeline_step[0]: {
                        "model": pipeline_step[1],
                        "input": last_name,
                    }
                }
            )
            last_name = pipeline_step[0]
        self._pipenet.steps = pipenet_steps

    def fit(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        skip_fit: Optional[List[str]] = None,
        return_intermediate: bool = False,
    ) -> Optional[Dict[str, Optional[Union[pd.Series, pd.DataFrame]]]]:
        """Train all models in the pipeline sequentially.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series used to train models.

        skip_fit: list, optional
            Models to skip training. This could be used when pipeline contains
            models that are already trained by the same time series, and re-
            training would be time consuming. It must be a list of strings
            where each element is a model name. Default: None.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        Returns
        -------
        dict, optional
            If return_intermediate=True, return intermediate results generated
            during training as a dictionary where keys are step names. If a
            step does not perform transformation or detection, the result of
            that step will be None.

        """
        self._update_internal_pipenet()
        return self._pipenet.fit(
            ts=ts, skip_fit=skip_fit, return_intermediate=return_intermediate
        )

    def detect(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        return_intermediate: bool = False,
        return_list: bool = False,
    ) -> Union[
        Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        Dict[
            str,
            Union[
                pd.Series,
                pd.DataFrame,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
                Dict[
                    str,
                    List[
                        Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]
                    ],
                ],
            ],
        ],
    ]:
        """Transform time series sequentially along pipeline, and detect
        anomalies with the last detector.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to detect anomalies from.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, list, or dict
            Detected anomalies.

            - If return_intermediate=False, return detected anomalies, i.e.
              result from last detector.
            - If return_intermediate=True, return results of all models in
              pipeline as a dict where each item represents the result of a
              model.
            - If return_list=False, result from a detector or an aggregator
              will be a binary pandas Series indicating normal/anomalous.
            - If return_list=True, result from a detector or an aggregator
              will be a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.


        """
        self._update_internal_pipenet()
        return self._pipenet.detect(
            ts=ts,
            return_intermediate=return_intermediate,
            return_list=return_list,
        )

    def transform(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        return_intermediate: bool = False,
    ) -> Union[
        Union[pd.Series, pd.DataFrame],
        Dict[str, Union[pd.Series, pd.DataFrame]],
    ]:
        """Transform time series sequentially along pipeline.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to be transformed

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, or dict
            Transformed time series.

            - If return_intermediate=False, return transformed series, i.e.
              result from last transformer;
            - If return_intermediate=True, return results of all models in
              pipeline as a dict where each item represents the result of a
              model.

        """
        self._update_internal_pipenet()
        return self._pipenet.transform(
            ts=ts, return_intermediate=return_intermediate
        )

    def fit_detect(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        skip_fit: Optional[List[str]] = None,
        return_intermediate: bool = False,
        return_list: bool = False,
    ) -> Union[
        Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        Dict[
            str,
            Union[
                pd.Series,
                pd.DataFrame,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
                Dict[
                    str,
                    List[
                        Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]
                    ],
                ],
            ],
        ],
    ]:
        """Train models in pipeline sequentially, transform time series along
        pipeline, and use the last detector to detect anomalies.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to detect anomalies from.

        skip_fit: list, optional
            Models to skip training. This could be used when pipeline contains
            models that are already trained by the same time series, and re-
            training would be time consuming. It must be a list of strings
            where each element is a model name. Default: None.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, list, or dict
            Detected anomalies.

            - If return_intermediate=False, return detected anomalies, i.e.
              result from last detector.
            - If return_intermediate=True, return results of all models in
              pipeline as a dict where each item represents the result of a
              model.
            - If return_list=False, result from a detector or an aggregator
              will be a binary pandas Series indicating normal/anomalous.
            - If return_list=True, result from a detector or an aggregator
              will be a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.


        """
        self._update_internal_pipenet()
        return self._pipenet.fit_detect(
            ts=ts,
            skip_fit=skip_fit,
            return_intermediate=return_intermediate,
            return_list=return_list,
        )

    def fit_transform(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        skip_fit: Optional[List[str]] = None,
        return_intermediate: bool = False,
    ) -> Union[
        Union[pd.Series, pd.DataFrame],
        Dict[str, Union[pd.Series, pd.DataFrame]],
    ]:
        """Train models in pipeline sequentially, and transform time series
        along pipeline.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to be transformed.

        skip_fit: list, optional
            Models to skip training. This could be used when pipeline contains
            models that are already trained by the same time series, and re-
            training would be time consuming. It must be a list of strings
            where each element is a model name. Default: None.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, or dict
            Transformed time series.

            - If return_intermediate=False, return transformed series, i.e.
              result from last transformer;
            - If return_intermediate=True, return results of all models in
              pipeline as a dict where each item represents the result of a
              model.

        """
        self._update_internal_pipenet()
        return self._pipenet.fit_transform(
            ts=ts, skip_fit=skip_fit, return_intermediate=return_intermediate
        )

    def score(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        anomaly_true: Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        scoring: str = "recall",
        **kwargs: Any
    ) -> Union[float, Dict[str, float]]:
        """Detect anomalies and score the results against true anomalies.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to detect anomalies from.

        anomaly_true: pandas Series or list
            True anomalies.

            - If pandas Series, it is treated as a series of binary labels.
            - If list, a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.

        scoring: str, optional
            Scoring function to use. Must be one of "recall", "precision",
            "f1", and "iou". See module `metrics` for more information.
            Default: "recall"

        **kwargs
            Optional parameters for scoring function. See module `metrics` for
            more information.

        Returns
        -------
        float
            Score of detection result.

        """
        if scoring == "recall":
            scoring_func = recall  # type: Callable
        elif scoring == "precision":
            scoring_func = precision
        elif scoring == "f1":
            scoring_func = f1_score
        elif scoring == "iou":
            scoring_func = iou
        else:
            raise ValueError(
                "Argument `scoring` must be one of 'recall', 'precision', "
                "'f1' and 'iou'."
            )
        if isinstance(anomaly_true, pd.Series):
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=False),
                **kwargs
            )
        else:
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=True),
                **kwargs
            )

    def get_params(self) -> Dict[str, Dict[str, Any]]:
        """Get parameters of models in pipeline.

        Returns
        -------
        dict
            A dictionary of model name and model parameters.

        """
        self._update_internal_pipenet()
        return self._pipenet.get_params()


class Pipenet:
    """A Pipenet object connects transformers, detectors and aggregators.

    Parameters
    ----------
    steps: dicts
        Components of the pipenet. Each key-value item represents a step (
        transformer, detector, or aggregator), where key is the unique name of
        the step and the value is a dict with the following key-value pairs:

            - input (str or list of str): Input to the model, which must be
              either 'original' (i.e. the input time series), or the name of
              a upstream component.
            - subset (str, list of str, or list of lists of str, optional): If
              a model does not use all series from an input component, use this
              field to specify which series should be included. If not given or
              "all", all series from the input component will be used.
            - model (object): A detector, transformer, or aggregator object.

    Attributes
    ----------
    steps_graph_: OrderedDict
        Order of steps to be executed. Keys are step names, values are 2-tuple
        (i, j) where i is the index of execution round and j is the the index
        within a round.

    final_step_: str
        Name of the final step to be executed. It is the single step in the
        last round of execution in attribute `steps_graph_`.

    Examples
    --------
    The following example show how to use a Pipenet to build a level shift
    detector with some basic transformers, detectors, and aggregator.

    >>> from adtk.detector import QuantileAD, ThresholdAD
    >>> from adtk.transformer import DoubleRollingAggregate
    >>> from adtk.aggregator import AndAggregator
    >>> from adtk.pipe import Pipenet
    >>> steps = {
            "diff_abs": {
                "input": "original",
                "model": DoubleRollingAggregate(
                    agg="median",
                    window=20,
                    center=True,
                    diff="l1",
                ),
            },
            "quantile_ad": {
                "input": "diff_abs",
                "model": QuantileAD(high=0.99, low=0),
            },
            "diff": {
                "input": "original",
                "model": DoubleRollingAggregate(
                    agg="median",
                    window=20,
                    center=True,
                    diff="diff",
                ),
            },
            "sign_check": {
                "input": "diff",
                "model": ThresholdAD(high=0.0, low=-float("inf")),
            },
            "and": {
                "model": AndAggregator(),
                "input": ["quantile_ad", "sign_check"],
            },
        }
    >>> myPipenet = Pipenet(steps)

    """

    def __init__(
        self, steps: Optional[Dict[str, Dict[str, Any]]] = None
    ) -> None:
        if steps is None:
            self.steps = dict()
        else:
            self.steps = steps
        self._validate()

    def _validate(self) -> None:
        """
        Check the following issues and raise error if found
        - steps is not a list of dict
        - some step does not have required keys
        - some step has invalid keys other than required and optional keys
        - some step's `name` is not str
        - some steps have duplicated name
        - some step uses reserved name
        - some step's `input` is not str or list of str
        - some step's `input` is not a step name or "original"
        - some step's `subset` (optional) is not a str ("all"),
            or a list of (str ("all") or list of str)
        - some step's `subset` does not have the same length as `input`
        - aggregator has a `subset` field
        - some step's `model` is not a valid ADTK model object
        - upstream of a transformer or a detector must be a transformer
        - upstream of aggregator must be a detector or an aggregator

        Create a directed graph of model execution flowchart, when following
        issues will be checked:
        - some step cannot be reached
        - more than one outlet steps found

        """

        # check if step is a dict
        if not isinstance(self.steps, dict):
            raise TypeError("`steps` must be a dict objects.")
        if not all([isinstance(value, dict) for value in self.steps.values()]):
            raise TypeError("Values of dict `steps` must be dict objects.")

        # check if each step has valid keys
        if not all(
            [
                {"model", "input"}
                <= step.keys()
                <= {"model", "input", "subset"}
                for step in self.steps.values()
            ]
        ):
            raise KeyError(
                "Each step must be a dict object with keys `model`, `input`, "
                "and `subset` (optional)."
            )

        # check if each step has valid name
        if not all(
            [isinstance(step_name, str) for step_name in self.steps.keys()]
        ):
            raise TypeError("Model name must be a string.")
        if any([step_name == "original" for step_name in self.steps.keys()]):
            raise ValueError(
                "'original' is a reserved name for original time series input "
                "and you may not use it as a model name."
            )

        # check if each step has valid input
        def islistofstr(li: Any) -> bool:
            if not isinstance(li, list):
                return False
            if not all([isinstance(x, str) for x in li]):
                return False
            return True

        if not all(
            [
                isinstance(step["input"], str) or islistofstr(step["input"])
                for step in self.steps.values()
            ]
        ):
            raise TypeError(
                "Field `input` must be a string or a list of strings."
            )
        name_set = set(self.steps.keys()).union({"original"})
        for step in self.steps.values():
            if (
                isinstance(step["input"], str)
                and (step["input"] not in name_set)
            ) or (
                isinstance(step["input"], list)
                and (not set(step["input"]) <= name_set)
            ):
                raise ValueError(
                    "Field `input` must be 'original' or name of a model."
                )

        # check if only one step is not used as input (so it is the output)
        for step in self.steps.values():
            if isinstance(step["input"], str):
                name_set = name_set - {step["input"]}
            else:
                name_set = name_set - set(step["input"])
        if len(name_set) > 1:
            raise ValueError(
                "Pipenet output is ambiguous: "
                "found more than one steps with no downstream step: {}.".format(
                    name_set
                )
            )

        # check if each step has a valid subset, or has no subset (then we
        # will add default one to it)
        for step_name, step in self.steps.items():
            if isinstance(step["input"], str):
                if "subset" not in step.keys():
                    pass
                elif not (
                    (
                        isinstance(step["subset"], str)
                        or islistofstr(step["subset"])
                    )
                ):
                    raise TypeError(
                        "Field `subset` at step '{}' is invalid.".format(
                            step_name
                        )
                    )
                elif isinstance(step["subset"], str) and (
                    step["subset"] != "all"
                ):
                    raise ValueError(
                        "A subset corresponding to an input source must be "
                        "'all' or a list of strings (even if there is only "
                        "one element)."
                    )
            else:
                if "subset" not in step.keys():
                    pass
                elif isinstance(step["subset"], str):
                    if step["subset"] == "all":
                        pass
                    else:
                        raise ValueError(
                            "Field `subset` at step '{}' is invalid.".format(
                                step_name
                            )
                        )
                elif isinstance(step["subset"], list):
                    if len(step["input"]) != len(step["subset"]):
                        raise ValueError(
                            "Fields `input` and `subset` are inconsistent at "
                            "step '{}'.".format(step_name)
                        )
                    for subset in step["subset"]:
                        if not (
                            (isinstance(subset, str) or islistofstr(subset))
                        ):
                            raise TypeError(
                                "Field `subset` at step '{}' is invalid.".format(
                                    step_name
                                )
                            )
                        if isinstance(subset, str) and (subset != "all"):
                            raise ValueError(
                                "A subset corresponding to an input source "
                                "must be 'all' or a list of strings (even if "
                                "there is only one element)."
                            )
                else:
                    raise TypeError(
                        "Field `subset` at step '{}' is invalid.".format(
                            step_name
                        )
                    )

        # check if each step has valid model
        if not all(
            [isinstance(step["model"], _Model) for step in self.steps.values()]
        ):
            raise ValueError(
                "Model must be a detector, transformer, or aggregator object."
            )

        # check:
        # 1. upstream of transformer and detector must be transformer, or input
        # 2. upstream of aggregator must be detector or aggregator
        for step_name, step in self.steps.items():
            if isinstance(step["model"], (_Detector, _Transformer)):
                if isinstance(step["input"], str):
                    if step["input"] == "original":
                        pass
                    elif not isinstance(
                        self.steps[step["input"]]["model"], _Transformer
                    ):
                        raise TypeError(
                            "Model in step '{}' cannot accept output from "
                            "step '{}'.".format(step_name, step["input"])
                        )
                else:
                    for input in step["input"]:
                        if input == "original":
                            pass
                        elif not isinstance(
                            self.steps[input]["model"], _Transformer
                        ):
                            raise TypeError(
                                "Model in step '{}' cannot accept output from "
                                "step '{}'.".format(step_name, input)
                            )
            elif isinstance(step["model"], _Aggregator):
                if isinstance(step["input"], str):
                    if (step["input"] == "original") or (
                        not isinstance(
                            self.steps[step["input"]]["model"],
                            (_Detector, _Aggregator),
                        )
                    ):
                        raise TypeError(
                            "Model in step '{}' cannot accept output from "
                            "step '{}'.".format(step_name, step["input"])
                        )
                else:
                    for input in step["input"]:
                        if (input == "original") or (
                            not isinstance(
                                self.steps[input]["model"],
                                (_Detector, _Aggregator),
                            )
                        ):
                            raise TypeError(
                                "Model in step '{}' cannot accept output from "
                                "step '{}'.".format(step_name, input)
                            )

        # sort out graph
        done_step = OrderedDict({"original": (0, 0)})
        counter = 0
        while True:
            counter += 1
            sub_counter = 0
            done_step_name_up_to_last_round = set(done_step.keys())
            for step_name, step in self.steps.items():
                if step_name in done_step_name_up_to_last_round:
                    continue
                if isinstance(step["input"], str):
                    if step["input"] in done_step_name_up_to_last_round:
                        done_step.update({step_name: (counter, sub_counter)})
                        sub_counter += 1
                else:
                    if set(step["input"]) <= done_step_name_up_to_last_round:
                        done_step.update({step_name: (counter, sub_counter)})
                        sub_counter += 1
            if len(done_step) == len(self.steps) + 1:
                break
            if sub_counter == 0:
                raise ValueError(
                    "The following step(s) cannot be reached: {}.".format(
                        set(self.steps.keys()) - set(done_step.keys())
                    )
                )
        self.steps_graph_ = done_step.copy()
        self.final_step_ = list(self.steps_graph_.keys())[-1]

    @staticmethod
    def _get_input(
        step: Dict[str, Any], results: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Given a step block and an intermediate results dict, get the input of
        this step based on fields `input` and `subset`.
        """
        if isinstance(step["model"], (_Detector, _Transformer)):
            if isinstance(step["input"], str):
                if ("subset" not in step.keys()) or (step["subset"] == "all"):
                    input = results[step["input"]]
                else:
                    if len(step["subset"]) == 1:
                        input = results[step["input"]][step["subset"][0]]
                    else:
                        input = results[step["input"]][step["subset"]]
            else:
                if ("subset" not in step.keys()) or (step["subset"] == "all"):
                    input = pd.concat(
                        [results[input_name] for input_name in step["input"]],
                        axis=1,
                    )
                else:
                    input = pd.concat(
                        [
                            results[input_name]
                            if subset == "all"
                            else (
                                results[input_name][subset[0]]
                                if len(subset) == 1
                                else results[input_name][subset]
                            )
                            for input_name, subset in zip(
                                step["input"], step["subset"]
                            )
                        ],
                        axis=1,
                    )
        elif isinstance(step["model"], _Aggregator):
            if isinstance(step["input"], str):
                input = {step["input"]: results[step["input"]]}
            else:
                input = {
                    input_name: results[input_name]
                    for input_name in step["input"]
                }
        return input

    def fit(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        skip_fit: Optional[List[str]] = None,
        return_intermediate: bool = False,
    ) -> Optional[Dict[str, Optional[Union[pd.Series, pd.DataFrame]]]]:
        """Train models in the pipenet.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series used to train models.

        skip_fit: list, optional
            Models to skip training. This could be used when pipenet contains
            models that are already trained by the same time series, and re-
            training would be time consuming. It must be a list of strings
            where each element is a model name. Default: None.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        Returns
        -------
        dict, optional
            If return_intermediate=True, return intermediate results generated
            during training as a dictionary where keys are step names. If a
            step does not perform transformation or detection, the result of
            that step will be None.

        """
        self._validate()

        if skip_fit is None:
            skip_fit = []
        if not isinstance(skip_fit, list):
            raise TypeError("Parameter `skip_fit` must be a list.")
        if not set(skip_fit) <= set(self.steps.keys()):
            raise ValueError("Name(s) in `skip_fit` is not valid model name.")

        # determine the step needing fit and/or predict
        need_fit = {
            step_name: (
                isinstance(step["model"], _TrainableModel)
                & (step_name not in skip_fit)
            )
            for step_name, step in self.steps.items()
        }
        need_predict = {step_name: False for step_name in self.steps.keys()}
        for step_name in list(self.steps_graph_.keys())[:0:-1]:
            if need_fit[step_name] or need_predict[step_name]:
                if isinstance(self.steps[step_name]["input"], str):
                    input = self.steps[step_name]["input"]
                    if input != "original":
                        need_predict[input] = True
                else:
                    for input in self.steps[step_name]["input"]:
                        if input != "original":
                            need_predict[input] = True

        # run fit or fit_predict
        results = {"original": ts.copy()}
        for step_name in list(self.steps_graph_.keys())[1:]:
            step = self.steps[step_name]
            if not (need_fit[step_name] or need_predict[step_name]):
                results.update({step_name: None})
                continue
            input = self._get_input(step, results)
            if need_fit[step_name] and (not need_predict[step_name]):
                step["model"].fit(input)
                results.update({step_name: None})
            elif (not need_fit[step_name]) and need_predict[step_name]:
                results.update({step_name: step["model"].predict(input)})
            else:
                results.update({step_name: step["model"].fit_predict(input)})

        # return intermediate results
        if return_intermediate:
            return results
        else:
            return None

    def _predict(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        fit: bool,
        detect: bool,
        skip_fit: Optional[List[str]],
        return_intermediate: bool,
        return_list: bool = False,
    ) -> Union[
        Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        Dict[
            str,
            Union[
                pd.Series,
                pd.DataFrame,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
                Dict[
                    str,
                    List[
                        Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]
                    ],
                ],
            ],
        ],
    ]:
        """
        Private method for detect, transform, fit_detect, fit_transform

        Parameters
        ----------
        fit: bool
            Whether this call is for fit_detect/fit_transform or
            detect/transform.

        detect: bool
            Whether this call is for detect/fit_detect or
            transform/fit_transform.

        Others:
            Same as higher-level calls

        Returns
        -------
            Same as higher-level calls

        """
        self._validate()

        if skip_fit is None:
            skip_fit = []
        if not isinstance(skip_fit, list):
            raise TypeError("Parameter `skip_fit` must be a list.")
        if not set(skip_fit) <= set(self.steps.keys()):
            raise ValueError("Name(s) in `skip_fit` is not valid model name.")

        last_step_name = list(self.steps_graph_.keys())[-1]

        if detect:
            if isinstance(self.steps[last_step_name]["model"], _Transformer):
                raise RuntimeError(
                    "This seems a transformation pipenet, "
                    "because model at the final step '{}' is a transformer. "
                    "Please use method `{}transform` instead.".format(
                        last_step_name, "fit_" if fit else ""
                    )
                )
        else:
            if isinstance(
                self.steps[last_step_name]["model"], (_Detector, _Aggregator)
            ):
                raise RuntimeError(
                    "This seems a detection pipenet, "
                    "because model at the final step '{}' is "
                    "either a detector or an aggregator. "
                    "Please use method `{}detect` instead.".format(
                        last_step_name, "fit_" if fit else ""
                    )
                )

        results = {
            "original": ts.copy()
        }  # type: Dict[str,Union[pd.Series,pd.DataFrame,List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]], Dict[str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]]]]
        for step_name in list(self.steps_graph_.keys())[1:]:
            step = self.steps[step_name]
            input = self._get_input(step, results)
            results.update(
                {
                    step_name: (
                        (
                            step["model"].fit_predict(
                                input, return_list=return_list
                            )
                            if isinstance(
                                step["model"],
                                (
                                    _TrainableUnivariateDetector,
                                    _TrainableMultivariateDetector,
                                ),
                            )
                            else step["model"].predict(
                                input, return_list=return_list
                            )
                        )
                        if isinstance(step["model"], _Detector)
                        else (
                            step["model"].fit_predict(input)
                            if isinstance(
                                step["model"],
                                (
                                    _TrainableUnivariateTransformer,
                                    _TrainableMultivariateTransformer,
                                ),
                            )
                            else step["model"].predict(input)
                        )
                    )
                    if fit and (step_name not in skip_fit)
                    else (
                        step["model"].predict(input, return_list=return_list)
                        if isinstance(step["model"], _Detector)
                        else step["model"].predict(input)
                    )
                }
            )

        if return_intermediate:
            return results
        else:
            return results[last_step_name]

    def detect(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        return_intermediate: bool = False,
        return_list: bool = False,
    ) -> Union[
        Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        Dict[
            str,
            Union[
                pd.Series,
                pd.DataFrame,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
                Dict[
                    str,
                    List[
                        Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]
                    ],
                ],
            ],
        ],
    ]:
        """Detect anomaly from time series using the pipenet.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to detect anomalies from.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, list, or dict
            Detected anomalies.

            - If return_intermediate=False, return detected anomalies, i.e.
              result from last detector.
            - If return_intermediate=True, return results of all models in
              pipenet as a dict where each item represents the result of a
              model.
            - If return_list=False, result from a detector or an aggregator
              will be a binary pandas Series indicating normal/anomalous.
            - If return_list=True, result from a detector or an aggregator
              will be a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.

        """
        return self._predict(
            ts,
            fit=False,
            detect=True,
            skip_fit=None,
            return_intermediate=return_intermediate,
            return_list=return_list,
        )

    def transform(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        return_intermediate: bool = False,
    ) -> Union[
        Union[pd.Series, pd.DataFrame],
        Dict[str, Union[pd.Series, pd.DataFrame]],
    ]:
        """Transform time series using the pipenet.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to be transformed.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, or dict
            Transformed time series.

            - If return_intermediate=False, return transformed series, i.e.
              result from last transformer;
            - If return_intermediate=True, return results of all models in
              pipnet as a dict where each item represents the result of a
              model.

        """
        return self._predict(
            ts,
            fit=False,
            detect=False,
            skip_fit=None,
            return_intermediate=return_intermediate,
        )

    def fit_detect(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        skip_fit: Optional[List[str]] = None,
        return_intermediate: bool = False,
        return_list: bool = False,
    ) -> Union[
        Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        Dict[
            str,
            Union[
                pd.Series,
                pd.DataFrame,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
                Dict[
                    str,
                    List[
                        Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]
                    ],
                ],
            ],
        ],
    ]:
        """Train models in the pipenet and detect anomaly with it.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to detect anomalies from.

        skip_fit: list, optional
            Models to skip training. This could be used when pipenet contains
            models that are already trained by the same time series, and re-
            training would be time consuming. It must be a list of strings
            where each element is a model name. Default: None.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        return_list: bool, optional
            Whether to return a list of anomalous events, or a binary series
            indicating normal/anomalous. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, list, or dict
            Detected anomalies.

            - If return_intermediate=False, return detected anomalies, i.e.
              result from last detector.
            - If return_intermediate=True, return results of all models in
              pipenet as a dict where each item represents the result of a
              model.
            - If return_list=False, result from a detector or an aggregator
              will be a binary pandas Series indicating normal/anomalous.
            - If return_list=True, result from a detector or an aggregator
              will be a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.

        """
        return self._predict(
            ts,
            fit=True,
            detect=True,
            skip_fit=skip_fit,
            return_intermediate=return_intermediate,
            return_list=return_list,
        )

    def fit_transform(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        skip_fit: Optional[List[str]] = None,
        return_intermediate: bool = False,
    ) -> Union[
        Union[pd.Series, pd.DataFrame],
        Dict[str, Union[pd.Series, pd.DataFrame]],
    ]:
        """Train models in the pipenet and transform time series with it.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to be transformed.

        skip_fit: list, optional
            Models to skip training. This could be used when pipenet contains
            models that are already trained by the same time series, and re-
            training would be time consuming. It must be a list of strings
            where each element is a model name. Default: None.

        return_intermediate: bool, optional
            Whether to return intermediate results. Default: False.

        Returns
        -------
        pandas Series, pandas DataFrame, or dict
            Transformed time series.

            - If return_intermediate=False, return transformed series, i.e.
              result from last transformer;
            - If return_intermediate=True, return results of all models in
              pipenet as a dict where each item represents the result of a
              model.


        """
        return self._predict(
            ts,
            fit=True,
            detect=False,
            skip_fit=skip_fit,
            return_intermediate=return_intermediate,
        )

    def score(
        self,
        ts: Union[pd.Series, pd.DataFrame],
        anomaly_true: Union[
            pd.Series,
            pd.DataFrame,
            List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
        scoring: str = "recall",
        **kwargs: Any
    ) -> Union[float, Dict[str, float]]:
        """Detect anomalies and score the results against true anomalies.

        Parameters
        ----------
        ts: pandas Series or DataFrame
            Time series to detect anomalies from.

        anomaly_true: Series, or a list of Timestamps or Timestamp tuple
            True anomalies.

            - If pandas Series, it is treated as a series of binary labels.
            - If list, a list of events where an event is a pandas Timestamp if
              it is instantaneous or a 2-tuple of pandas Timestamps if it is a
              closed time interval.

        scoring: str, optional
            Scoring function to use. Must be one of "recall", "precision",
            "f1", and "iou". See module `metrics` for more information.
            Default: "recall"

        **kwargs
            Optional parameters for scoring function. See module `metrics` for
            more information.

        Returns
        -------
        float
            Score of detection result.

        """
        if scoring == "recall":
            scoring_func = recall  # type: Callable
        elif scoring == "precision":
            scoring_func = precision
        elif scoring == "f1":
            scoring_func = f1_score
        elif scoring == "iou":
            scoring_func = iou
        else:
            raise ValueError(
                "Argument `scoring` must be one of 'recall', 'precision', "
                "'f1' and 'iou'."
            )
        if isinstance(anomaly_true, pd.Series):
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=False),
                **kwargs
            )
        else:
            return scoring_func(
                y_true=anomaly_true,
                y_pred=self.detect(ts, return_list=True),
                **kwargs
            )

    def get_params(self) -> Dict[str, Dict[str, Any]]:
        """Get parameters of models in pipenet.

        Returns
        -------
        dict
            A dictionary of model name and model parameters.

        """
        return {
            step_name: step["model"].get_params()
            for step_name, step in self.steps.items()
        }

    def summary(self) -> None:
        """Print a summary of the pipenet."""
        df = pd.DataFrame(columns=["name", "model", "input", "subset"])
        for step_name in self.steps_graph_.keys():
            if step_name == "original":
                continue
            df = df.append(
                {
                    "name": step_name,
                    "model": self.steps[step_name]["model"].__class__.__name__,
                    "input": self.steps[step_name]["input"],
                    "subset": (
                        self.steps[step_name]["subset"]
                        if "subset" in self.steps[step_name].keys()
                        else "all"
                    ),
                },
                ignore_index=True,
            )
        print(tabulate(df, headers="keys", tablefmt="simple", showindex=False))

    def plot_flowchart(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[float, float]] = None,
        radius: float = 1.0,
    ) -> plt.Axes:  # pragma: no cover
        """Plot flowchart of this pipenet.

        Parameters
        ----------
        ax: matplotlib axes object, optional
            Axes to plot at. If not given, the method will create a matplotlib
            figure and axes. Default: None.

        figsize: tuple, optional
            Width and height of the figure to plot at. Only to be used if `ax`
            is not given. Default: None.

        radius: float, optional
            Relative size of components in the chart. Default: 1.0.

        Returns
        -------
        matplotlib axes object
            Axes where the flowchart is plotted.

        """
        self._validate()

        radius = radius * 0.1

        # create empty plot
        if (ax is None) and (figsize is None):
            figsize = (14, 16)
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)

        # get coordinate of components
        layers = []  # type: List
        for graph_step, values in self.steps_graph_.items():
            if values[1] == 0:
                layers.append([])
            layers[-1].append(graph_step)
        n_layer = len(layers)
        max_n_comp = 0
        coord = dict()
        for layer in layers:
            n_comp = len(layer)
            max_n_comp = max(max_n_comp, n_comp)
            for comp in layer:
                x = self.steps_graph_[comp][0]
                y = self.steps_graph_[comp][1] - (n_comp - 1) / 2
                coord[comp] = (x, y)

        # plot connection lines, and gather patches
        io_patches = []  # type: List
        detector_patches = []  # type: List
        transformer_patches = []  # type: List
        aggregator_patches = []  # type: List
        for step_name, step in self.steps.items():
            end_coord = coord[step_name]
            if isinstance(step["model"], _Detector):
                detector_patches.append(Circle(xy=end_coord, radius=radius))
            elif isinstance(step["model"], _Transformer):
                transformer_patches.append(Circle(xy=end_coord, radius=radius))
            elif isinstance(step["model"], _Aggregator):
                aggregator_patches.append(Circle(xy=end_coord, radius=radius))
            if isinstance(step["input"], str):
                input = step["input"]
                start_coord = coord[input]
                if ("subset" not in step.keys()) or (step["subset"] == "all"):
                    linestyle = "-"
                else:
                    linestyle = "--"
                plt.plot(
                    [start_coord[0], end_coord[0]],
                    [start_coord[1], end_coord[1]],
                    color="k",
                    linestyle=linestyle,
                    zorder=1,
                )
            else:
                for counter, input in enumerate(step["input"]):
                    start_coord = coord[input]
                    if (
                        ("subset" not in step.keys())
                        or (step["subset"] == "all")
                        or (step["subset"][counter] == "all")
                    ):
                        linestyle = "-"
                    else:
                        linestyle = "--"
                    plt.plot(
                        [start_coord[0], end_coord[0]],
                        [start_coord[1], end_coord[1]],
                        color="k",
                        linestyle=linestyle,
                        zorder=1,
                    )
        plt.plot([n_layer - 1, n_layer], [0, 0], color="k", zorder=1)
        io_patches.append(Circle(xy=(0, 0), radius=radius))
        io_patches.append(Circle(xy=(n_layer, 0), radius=radius))

        # draw components
        ax.add_collection(PatchCollection(io_patches, zorder=2, color="y"))
        ax.add_collection(
            PatchCollection(detector_patches, zorder=2, color="g")
        )
        ax.add_collection(
            PatchCollection(transformer_patches, zorder=2, color="c")
        )
        ax.add_collection(
            PatchCollection(aggregator_patches, zorder=2, color="m")
        )

        # label components with step names
        for key, value in coord.items():
            ax.annotate(
                key, (value[0] - radius, value[1]), zorder=3, color="k"
            )
        ax.annotate("result", (n_layer - radius, 0), zorder=3, color="k")

        # add legend
        plt.legend(
            handles=[
                Line2D([], [], color="k", label="full connection"),
                Line2D(
                    [],
                    [],
                    color="k",
                    linestyle="--",
                    label="partial connection",
                ),
                Circle([], radius=radius, color="y", label="input/output"),
                Circle([], radius=radius, color="g", label="detector"),
                Circle([], radius=radius, color="c", label="transformer"),
                Circle([], radius=radius, color="m", label="aggregator"),
            ],
            bbox_to_anchor=(1, 1),
            loc=2,
            borderaxespad=0.0,
        )

        # clean up axes
        ax.set_xlim([-radius - 0.1, n_layer + radius + 0.1])
        ax.set_ylim(
            [
                -(max_n_comp - 1) / 2 - radius - 0.1,
                (max_n_comp - 1) / 2 + radius + 0.1,
            ]
        )
        ax.set_axis_off()
        ax.set_aspect(1)

        return ax



"""
Module of metrics that measure the quality of detection results against true
anomalies.
"""

from ._metrics import f1_score, iou, precision, recall

__all__ = ["recall", "precision", "f1_score", "iou"]



from typing import Union

import pandas as pd

from ._base import (
    _NonTrainableMultivariateModel,
    _NonTrainableUnivariateModel,
    _TrainableMultivariateModel,
    _TrainableUnivariateModel,
)


class _NonTrainableUnivariateTransformer(_NonTrainableUnivariateModel):
    def predict(
        self, ts: Union[pd.Series, pd.DataFrame]
    ) -> Union[pd.Series, pd.DataFrame]:
        """Transform time series.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to be transformed. If a DataFrame with k columns, it is
            treated as k independent univariate time series and the transformer
            will be applied to each univariate series independently.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            Transformed time series.

        """
        return self._predict(ts)

    transform = predict


class _TrainableUnivariateTransformer(_TrainableUnivariateModel):
    def fit(self, ts: Union[pd.Series, pd.DataFrame]) -> None:
        """Train the transformer with given time series.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to be used to train the transformer.
            If a DataFrame with k columns, k univariate transformers will be
            trained independently.

        """
        self._fit(ts)

    def predict(
        self, ts: Union[pd.Series, pd.DataFrame]
    ) -> Union[pd.Series, pd.DataFrame]:
        """Transform time series.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to be transformed. If a DataFrame with k columns, it is
            treated as k independent univariate time series.

            - If the transformer was trained with a Series, the transformer
              will be applied to each univariate series independently;
            - If the transformer was trained with a DataFrame, i.e. the
              transformer is essentially k transformers, those transformers
              will be applied to each univariate series respectively.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            Transformed time series.

        """
        return self._predict(ts)

    def fit_predict(
        self, ts: Union[pd.Series, pd.DataFrame]
    ) -> Union[pd.Series, pd.DataFrame]:
        """Train the transformer, and tranform the time series used for
        training.

        Parameters
        ----------
        ts: pandas.Series or pandas.DataFrame
            Time series to be used for training and be transformed.
            If a DataFrame with k columns, it is treated as k independent
            univariate time series, and k univariate transformers will be
            trained and applied to each series independently.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            Transformed time series.

        """
        self.fit(ts)
        return self.predict(ts)

    transform = predict
    fit_transform = fit_predict


class _NonTrainableMultivariateTransformer(_NonTrainableMultivariateModel):
    def predict(self, df: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """Transform time series.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to be transformed.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            Transformed time series.

        """
        return self._predict(df)

    transform = predict


class _TrainableMultivariateTransformer(_TrainableMultivariateModel):
    def fit(self, df: pd.DataFrame) -> None:
        """Train the transformer with given time series.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to be used to train the transformer.

        """
        self._fit(df)

    def predict(self, df: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """Transform time series.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to be transformed.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            Transformed time series.

        """
        return self._predict(df)

    def fit_predict(self, df: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """Train the transformer, and tranform the time series used for
        training.

        Parameters
        ----------
        df: pandas.DataFrame
            Time series to be used for training and be transformed.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            Transformed time series.

        """
        self.fit(df)
        return self.predict(df)

    transform = predict
    fit_transform = fit_predict



"""Module of aggregators.

An aggregator combines multiple lists of anomalies into one.

"""
from typing import Dict, Optional

from .._aggregator_base import _Aggregator
from .._utils import _get_all_subclasses_from_superclass
from ._aggregator import AndAggregator, CustomizedAggregator, OrAggregator


def print_all_models() -> None:
    """
    Print description of every model in this module.
    """
    model_desc = _get_all_subclasses_from_superclass(
        _Aggregator
    )  # type: Dict[str, Optional[str]]
    for key, value in model_desc.items():
        print("-" * 80)
        print(key)
        print(value)


__all__ = [
    "OrAggregator",
    "AndAggregator",
    "CustomizedAggregator",
    "print_all_models",
]



"""
We don't typing the visualization module because there are a lot recursion on
nested tree structure which would be messy if we type rigorously."""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.plotting import register_matplotlib_converters

from ..data import to_events, to_labels, validate_events

register_matplotlib_converters()


def plot(
    ts=None,
    anomaly=None,
    curve_group="each",
    ts_linewidth=0.5,
    ts_color=None,
    ts_alpha=1.0,
    ts_marker=".",
    ts_markersize=2,
    match_curve_name=True,
    anomaly_tag="span",
    anomaly_color=None,
    anomaly_alpha=0.3,
    anomaly_marker="o",
    anomaly_markersize=4,
    freq_as_period=True,
    axes=None,
    figsize=None,
    legend=True,
):
    """Plot time series and/or anomalies.

    Parameters
    ----------
    ts: pandas Series or DataFrame, optional
        Time series to plot.

    anomaly: list, pandas Series, DataFrame, or (nested) dict of them, optional
        Anomalies to plot.

        - If list, a list of anomalous events (pandas Timestamp for an
          instantaneous event or 2-tuple of pandas Timestamps for an interval);
        - If pandas Series, a binary series indicating normal/anomalous;
        - If pandas DataFrame, each column is treated independently as a binary
          Series.
        - If (nested) dict, every leaf node (list, Series, or DataFrame) is
          treated independently as above.

    curve_group: str or list, optional
        Groups of curves to be drawn at same plots.

        - If str, 'each' means every dimension is drawn in a separated plot,
          'all' means all dimensions are drawn in the same plot.
        - If list, each element corresponds to a subplot, which is the name of
          time series to plot in this subplot, or a list of names. For example,
          ["A", ("B", "C")] means two subplots, where the first one contain
          series A, while the second one contains series B and C.

        Default: 'each'.

    ts_linewidth: float or dict, optional
        Line width of each time series curve.

        - If float, all curves have the same line width.
        - If dict, the key is series name, the value is line width of that
          series.

        Default: 0.5.

    ts_color: str or dict, optional
        Color of each time series curve.

        - If str, all curves have the same color.
        - If dict, the key is series name, the value is color of that series.
        - If None, color will be assigned automatically.

        Default: None.

    ts_alpha: float or dict, optional
        Opacity of each time series curve.

        - If float, all curves have the same opacity.
        - If dict, the key is series name, the value is opacity of that series.

        Default: 1.0.

    ts_marker: str or dict, optional
        Marker type of each time series curve.

        - If str, all curves have the same marker type.
        - If dict, the key is series name, the value is marker type of that
          series.

        Default: ".".

    ts_markersize: int or dict, optional
        Marker size of each time series curve.

        - If int, all curves have the same marker size.
        - If dict, the key is series name, the value is marker size of that
          series.

        Default: 2.

    match_curve_name: bool, optional
        Whether to plot anomaly with corresponding curve by matching series
        names. If False, plot anomaly with all curves.
        Default: True.

    anomaly_tag: str, or (nested) dict, optional
        Plot anomaly as horizontal spans or markers on curves.

        - If str, either 'span' or 'marker', all anomalies are marked with the
          same type of tag.
        - If (nested) dict, it must have a tree structure identical to or
          smaller than that of (nested) dict argument `anomaly`, which can
          define tags for all leaf nodes in `anomaly`.

        Default: "span".

    anomaly_color: str, or (nested) dict, optional
        Color of each anomaly tag.

        - If str, all anomalies are marked with the same color.
        - If (nested) dict, it must have a tree structure identical to or
          smaller than that of (nested) dict argument `anomaly`, which can
          define colors for all leaf nodes in `anomaly`.
        - If None, color will be assigned automatically.

        Default: None.

    anomaly_alpha: float, or (nested) dict, optional
        Opacity of each anomaly tag. Only used for anomaly drawn as horizontal
        spans.

        - If float, all anomalies are marked with the same opacity.
        - If (nested) dict, it must have a tree structure identical to or
          smaller than that of (nested) dict argument `anomaly`, which can
          define opacity for all leaf nodes in `anomaly`.

        Default: 0.3.

    anomaly_marker: str, or (nested) dict, optional
        Marker type of each anomaly marker. Only used for anomaly drawn as
        markers on curves.

        - If str, all anomalies are marked with the same type of marker.
        - If (nested) dict, it must have a tree structure identical to or
          smaller than that of (nested) dict argument `anomaly`, which can
          define marker types for all leaf nodes in `anomaly`.

        Default: "o".

    anomaly_markersize: int, or (nested) dict, optional
        Marker size of each anomaly marker. Only used for anomaly drawn as
        markers on curves.

        - If int, all anomalies are marked with the same size of marker.
        - If (nested) dict, it must have a tree structure identical to or
          smaller than that of (nested) dict argument `anomaly`, which can
          define marker sizes for all leaf nodes in `anomaly`.

        Default: 4.

    freq_as_period: bool, optional
        Whether to regard time index with regular frequency (i.e. attribute
        `freq` of time index is not None) as time intervals. Only used when
        anomaly is given as binary series.

        For example, DatetimeIndex(['2017-01-01', '2017-01-02', '2017-01-03',
        '2017-01-04', '2017-01-05'], dtype='datetime64[ns]', freq='D') has
        daily frequency. If freq_as_period=True, each time point in the index
        represents that day (24 hours). Otherwsie, each time point represents
        the instantaneous time instance of 00:00:00 on that day.

        Default: True.

    axes: matplotlib Axes object, or array of Axes objects, optional
        Axes to plot at. The number of Axes objects should be equal to the
        number of plots. If not specified, figure axes will be automatically
        generated. Default: None.

    figsize: tuple, optional
        Size of the figure. If not specified, the size of each subplot is 16 x
        4. Default: None.

    legend: bool, optional
        Whether to show legend in the plot. Default: True.

    Returns
    --------
    matplotlib Axes object or array of Axes objects
        Axes where the plot(s) is drawn.

    """
    # setup style
    plt.style.use("seaborn-whitegrid")

    # initialize color generator
    color_generator = ColorGenerator()

    # plot time series
    if ts is not None:
        # type check for ts
        if isinstance(ts, pd.Series):
            if ts.name is None:
                df = ts.to_frame("Time Series")
            else:
                df = ts.to_frame()
        elif isinstance(ts, pd.DataFrame):
            df = ts.copy()
        else:
            raise TypeError(
                "Argument `ts` must be a pandas Series or DataFrame."
            )

        # check series index
        if not isinstance(df.index, pd.DatetimeIndex):
            raise TypeError(
                "Index of the input time series must be a pandas "
                "DatetimeIndex object."
            )

        # check duplicated column names
        if df.columns.duplicated().any():
            raise ValueError("Input DataFrame must have unique column names.")

        # set up curve groups
        if curve_group == "each":
            curve_group = list(df.columns)
        elif curve_group == "all":
            curve_group = [tuple(df.columns)]

        # validate curve groups
        curve2axes = _validate_curve_group(df, curve_group)

        # set up default figure size
        if figsize is None:
            figsize = (16, 4 * len(curve_group))

        # setup axes
        if axes is None:
            _, axes = plt.subplots(
                nrows=len(curve_group), figsize=figsize, sharex=True
            )
        if not isinstance(axes, (list, np.ndarray)):
            axes = [axes]
        for ax in axes:
            ax.xaxis_date()

        # expand ts properties to a dict, if not yet
        ts_color = _assign_properties(ts_color, df)
        ts_linewidth = _assign_properties(ts_linewidth, df, 0.5)
        ts_marker = _assign_properties(ts_marker, df, ".")
        ts_markersize = _assign_properties(ts_markersize, df, 2)
        ts_alpha = _assign_properties(ts_alpha, df, 1.0)

        # plot curves
        _plot_curve(
            df,
            axes,
            curve2axes,
            ts_color,
            ts_linewidth,
            ts_marker,
            ts_markersize,
            ts_alpha,
            color_generator,
        )
    else:  # no time series, just event
        df = pd.DataFrame(dtype=int)
        curve2axes = dict()

        # never try to match curve name, because there is no curve anyway
        match_curve_name = False

        # never try to plot on curve, because there is no curve anyway
        anomaly_tag = "span"

        # setup figure
        if figsize is None:
            figsize = (16, 4)

        # setup axes
        if axes is None:
            _, axes = plt.subplots(figsize=figsize)
        if not isinstance(axes, (list, np.ndarray)):
            axes = [axes]
        for ax in axes:
            ax.xaxis_date()

    # plot anomaly
    if anomaly is not None:
        # validate anomaly
        _validate_anomaly(anomaly)
        # this is for showing a legend even if the series does not have a key
        if isinstance(anomaly, (list, pd.Series)):
            anomaly = {"Anomaly": anomaly}

        # expand tree struct of anomaly properties to match that of `anomaly`
        anomaly_tag = _assign_properties(anomaly_tag, anomaly, "span")
        anomaly_color = _assign_properties(anomaly_color, anomaly)
        anomaly_alpha = _assign_properties(anomaly_alpha, anomaly, 0.3)
        anomaly_marker = _assign_properties(anomaly_marker, anomaly, "o")
        anomaly_markersize = _assign_properties(anomaly_markersize, anomaly, 4)

        # plot anomalies
        _plot_anomaly(
            anomaly,
            axes,
            df,
            curve2axes,
            anomaly_tag,
            anomaly_color,
            anomaly_marker,
            anomaly_markersize,
            anomaly_alpha,
            match_curve_name,
            freq_as_period,
            color_generator,
        )

    # display legend
    if legend and ((ts is not None) or (anomaly is not None)):
        for ax in axes:
            ax.legend()

    return axes


def _validate_curve_group(df, curve_group):
    "Validate curve group, and return inverse map."
    curve2group = {col: set() for col in df.columns}
    for ind, group in enumerate(curve_group):
        if not isinstance(
            group, (list, tuple)
        ):  # this group has a single curve
            if group in set(df.columns):
                curve2group[group].add(ind)
            else:
                raise ValueError(
                    "{} is not a seriers in input `ts`.".format(group)
                )
        else:
            for curve in group:
                if curve in set(df.columns):
                    curve2group[curve].add(ind)
                else:
                    raise ValueError(
                        "{} is not a seriers in input `ts`.".format(curve)
                    )
    return curve2group


def _plot_curve(
    df,
    axes,
    curve2axes,
    ts_color,
    ts_linewidth,
    ts_marker,
    ts_markersize,
    ts_alpha,
    color_generator,
):
    "Plot all curves"
    for col, axes_inds in curve2axes.items():
        color = color_generator.emit(ts_color[col])
        for axes_ind in axes_inds:
            # df[col].plot(
            #     ax=axes[axes_ind],
            #     color=color,
            #     linewidth=ts_linewidth[col],
            #     marker=ts_marker[col],
            #     markersize=ts_markersize[col],
            #     alpha=ts_alpha[col],
            #     label=str(col),
            # )
            axes[axes_ind].plot_date(
                df.index,
                df[col],
                fmt="-",
                color=color,
                linewidth=ts_linewidth[col],
                marker=ts_marker[col],
                markersize=ts_markersize[col],
                alpha=ts_alpha[col],
                label=str(col),
            )


def _plot_anomaly(
    anomaly,
    axes,
    df,
    curve2axes,
    anomaly_tag,
    anomaly_color,
    anomaly_marker,
    anomaly_markersize,
    anomaly_alpha,
    match_curve_name,
    freq_as_period,
    color_generator,
    anomaly_name=None,
    anomaly_label=None,
):
    if isinstance(anomaly, (list, pd.Series)):
        anomaly_color = color_generator.emit(anomaly_color)
        if anomaly_tag == "span":
            # turn anomaly into list, if not yet
            if isinstance(anomaly, pd.Series):
                anomaly = to_events(anomaly, freq_as_period=freq_as_period)
            anomaly = validate_events(anomaly, point_as_interval=True)
            if match_curve_name and (
                anomaly_name in set(df.columns)
            ):  # match found, plot on it
                for axes_ind in curve2axes[anomaly_name]:
                    _add_anomaly_list_to_axes(
                        anomaly,
                        axes[axes_ind],
                        anomaly_color,
                        anomaly_alpha,
                        (
                            anomaly_label
                            if (anomaly_label != anomaly_name)
                            else "Anomaly - {}".format(anomaly_name)
                        ),
                    )
            else:  # not match found or don't match, plot on all
                for ax in axes:
                    _add_anomaly_list_to_axes(
                        anomaly,
                        ax,
                        anomaly_color,
                        anomaly_alpha,
                        (
                            "Anomaly - {}".format(anomaly_name)
                            if (
                                (anomaly_label == anomaly_name)
                                and (anomaly_name in set(df.columns))
                            )
                            else anomaly_label
                        ),
                    )
        elif anomaly_tag == "marker":
            # turn anomaly into binary series, if not yet
            if isinstance(anomaly, list):
                anomaly = to_labels(
                    anomaly, df.index, freq_as_period=freq_as_period
                )
            else:
                try:
                    pd.testing.assert_index_equal(
                        anomaly.index, df.index, check_names=False
                    )
                except AssertionError:
                    raise ValueError(
                        "Series index in argument `anomaly` must be the same "
                        "as the input time series."
                    )
            if match_curve_name and (
                anomaly_name in set(df.columns)
            ):  # match found, plot on it
                for axes_ind in curve2axes[anomaly_name]:
                    _add_anomaly_series_to_curve(
                        anomaly,
                        axes[axes_ind],
                        df[anomaly_name],
                        anomaly_color,
                        anomaly_marker,
                        anomaly_markersize,
                        (
                            anomaly_label
                            if (anomaly_label != anomaly_name)
                            else "Anomaly - {}".format(anomaly_name)
                        ),
                    )
            else:  # not match found or don't match, plot on all
                # hasLegend is an auxilary variable to make sure an anomaly
                # series only appears once in legend in an axes
                hasLegend = [False] * len(axes)
                for curve, axes_inds in curve2axes.items():
                    for axes_ind in axes_inds:
                        _add_anomaly_series_to_curve(
                            anomaly,
                            axes[axes_ind],
                            df[curve],
                            anomaly_color,
                            anomaly_marker,
                            anomaly_markersize,
                            (
                                "Anomaly - {}".format(anomaly_name)
                                if (
                                    (anomaly_label == anomaly_name)
                                    and (anomaly_name in set(df.columns))
                                )
                                else anomaly_label
                            )
                            if not hasLegend[axes_ind]
                            else None,
                        )
                        hasLegend[axes_ind] = True
        else:
            raise ValueError(
                "An anomaly tag must be either 'span' or 'marker'."
            )
    elif isinstance(anomaly, (pd.DataFrame, dict)):
        for key in (
            anomaly.columns
            if isinstance(anomaly, pd.DataFrame)
            else anomaly.keys()
        ):
            _plot_anomaly(
                anomaly[key],
                axes,
                df,
                curve2axes,
                anomaly_tag[key],
                anomaly_color[key],
                anomaly_marker[key],
                anomaly_markersize[key],
                anomaly_alpha[key],
                match_curve_name,
                freq_as_period,
                color_generator,
                anomaly_name=key,
                anomaly_label=(
                    "{} - {}".format(anomaly_name, key)
                    if (anomaly_name is not None)
                    else key
                ),
            )
    else:
        raise TypeError(
            "Argument `anomaly` must be a list, pandas Series, DataFrame, or "
            "a (nested) dict of them."
        )


def _add_anomaly_list_to_axes(
    anomaly, ax, anomaly_color, anomaly_alpha, anomaly_label
):
    "Add a list of anomalous event to an axes as spans"
    for i, event in enumerate(anomaly):
        ax.axvspan(
            xmin=event[0],
            xmax=event[1],
            color=anomaly_color,
            alpha=anomaly_alpha,
            label=(anomaly_label if i == 0 else None),
        )


def _add_anomaly_series_to_curve(
    anomaly,
    ax,
    s,
    anomaly_color,
    anomaly_marker,
    anomaly_markersize,
    anomaly_label,
):
    "Add anomalies represented by a binary series as markers on a curve"
    anomaly_curve = s.loc[anomaly == 1]
    # anomaly_curve.plot(
    #     ax=ax,
    #     linewidth=0,
    #     marker=anomaly_marker,
    #     markersize=anomaly_markersize,
    #     color=anomaly_color,
    #     label=anomaly_label,
    # )
    ax.plot_date(
        anomaly_curve.index,
        anomaly_curve,
        fmt="-",
        linewidth=0,
        marker=anomaly_marker,
        markersize=anomaly_markersize,
        color=anomaly_color,
        label=anomaly_label,
    )


def _validate_anomaly(anomaly):
    "Validate argument `anomaly`."
    if isinstance(anomaly, (list, pd.Series)):
        pass
    elif isinstance(anomaly, pd.DataFrame):
        if anomaly.columns.duplicated().any():
            raise ValueError(
                "DataFrame in argument `anomaly` must have unique column names."
            )
    elif isinstance(anomaly, dict):
        for _, value in anomaly.items():
            _validate_anomaly(value)
    else:
        raise TypeError(
            "Argument `anomaly` must be a list, pandas Series, DataFrame, or "
            "a (nested) dict of them."
        )


def _assign_properties(prop, anomaly, default=None):
    "Expand the tree structure of `prop` to that of `anomaly`"
    if (not isinstance(prop, dict)) and isinstance(
        anomaly, (dict, pd.DataFrame)
    ):
        return {
            key: _assign_properties(prop, anomaly[key])
            for key in (
                anomaly.keys()
                if isinstance(anomaly, dict)
                else anomaly.columns
            )
        }
    elif (not isinstance(prop, dict)) and (
        not isinstance(anomaly, (dict, pd.DataFrame))
    ):
        return prop
    elif isinstance(prop, dict) and (
        not isinstance(anomaly, (dict, pd.DataFrame))
    ):
        raise ValueError("Property dict and anomaly dict are inconsistent.")
    else:  # isinstance(prop, dict) & isinstance(anomaly, (dict, pd.DataFrame))
        if set(prop.keys()) <= set(
            anomaly.keys() if isinstance(anomaly, dict) else anomaly.columns
        ):
            return {
                key: _assign_properties(
                    (prop[key] if (key in prop.keys()) else default),
                    anomaly[key],
                )
                for key in (
                    anomaly.keys()
                    if isinstance(anomaly, dict)
                    else anomaly.columns
                )
            }
        else:
            raise ValueError(
                "Property dict and anomaly dict are inconsistent."
            )


class ColorGenerator:
    """
    Generate color
    """

    def __init__(self):
        self.latest_auto_color = -1

    def emit(self, color=None):
        if color is not None:
            return color
        else:
            self.latest_auto_color += 1
            return "C{}".format(self.latest_auto_color)
